@@ -1,4 +1,4 @@
-# Copyright 2009-2015 MongoDB, Inc.
+# Copyright 2009-present MongoDB, Inc.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -14,11 +14,12 @@
 
 """Tools for representing files stored in GridFS."""
 import datetime
+import hashlib
+import io
 import math
 import os
 
-from hashlib import md5
-
+from bson.int64 import Int64
 from bson.son import SON
 from bson.binary import Binary
 from bson.objectid import ObjectId
@@ -28,7 +29,9 @@ from pymongo import ASCENDING
 from pymongo.collection import Collection
 from pymongo.cursor import Cursor
 from pymongo.errors import (ConfigurationError,
+                            CursorNotFound,
                             DuplicateKeyError,
+                            InvalidOperation,
                             OperationFailure)
 from pymongo.read_preferences import ReadPreference
 
@@ -97,10 +100,23 @@ def _grid_out_property(field_name, docstring):
     return property(getter, doc=docstring)
 
 
+def _clear_entity_type_registry(entity, **kwargs):
+    """Clear the given database/collection object's type registry."""
+    codecopts = entity.codec_options.with_options(type_registry=None)
+    return entity.with_options(codec_options=codecopts, **kwargs)
+
+
+def _disallow_transactions(session):
+    if session and session.in_transaction:
+        raise InvalidOperation(
+            'GridFS does not support multi-document transactions')
+
+
 class GridIn(object):
     """Class to write data to GridFS.
     """
-    def __init__(self, root_collection, **kwargs):
+    def __init__(
+            self, root_collection, session=None, disable_md5=False, **kwargs):
         """Write a file to GridFS
 
         Application developers should generally not need to
@@ -136,8 +152,18 @@ class GridIn(object):
 
         :Parameters:
           - `root_collection`: root collection to write to
+          - `session` (optional): a
+            :class:`~pymongo.client_session.ClientSession` to use for all
+            commands
+          - `disable_md5` (optional): When True, an MD5 checksum will not be
+            computed for the uploaded file. Useful in environments where
+            MD5 cannot be used for regulatory or other reasons. Defaults to
+            False.
           - `**kwargs` (optional): file level options (see above)
 
+        .. versionchanged:: 3.6
+           Added ``session`` parameter.
+
         .. versionchanged:: 3.0
            `root_collection` must use an acknowledged
            :attr:`~pymongo.collection.Collection.write_concern`
@@ -146,10 +172,10 @@ class GridIn(object):
             raise TypeError("root_collection must be an "
                             "instance of Collection")
 
-        # With w=0, 'filemd5' might run before the final chunks are written.
         if not root_collection.write_concern.acknowledged:
             raise ConfigurationError('root_collection must use '
                                      'acknowledged write_concern')
+        _disallow_transactions(session)
 
         # Handle alternative naming
         if "content_type" in kwargs:
@@ -157,13 +183,15 @@ class GridIn(object):
         if "chunk_size" in kwargs:
             kwargs["chunkSize"] = kwargs.pop("chunk_size")
 
-        coll = root_collection.with_options(
-            read_preference=ReadPreference.PRIMARY)
+        coll = _clear_entity_type_registry(
+            root_collection, read_preference=ReadPreference.PRIMARY)
 
-        kwargs['md5'] = md5()
+        if not disable_md5:
+            kwargs["md5"] = hashlib.md5()
         # Defaults
         kwargs["_id"] = kwargs.get("_id", ObjectId())
         kwargs["chunkSize"] = kwargs.get("chunkSize", DEFAULT_CHUNK_SIZE)
+        object.__setattr__(self, "_session", session)
         object.__setattr__(self, "_coll", coll)
         object.__setattr__(self, "_chunks", coll.chunks)
         object.__setattr__(self, "_file", kwargs)
@@ -174,17 +202,20 @@ class GridIn(object):
         object.__setattr__(self, "_ensured_index", False)
 
     def __create_index(self, collection, index_key, unique):
-        doc = collection.find_one(projection={"_id": 1})
+        doc = collection.find_one(projection={"_id": 1}, session=self._session)
         if doc is None:
             try:
-                index_keys =[index_spec['key'] for index_spec in collection.list_indexes()]
+                index_keys = [index_spec['key'] for index_spec in
+                              collection.list_indexes(session=self._session)]
             except OperationFailure:
                 index_keys = []
             if index_key not in index_keys:
-                collection.create_index(index_key.items(), unique=unique)
+                collection.create_index(
+                    index_key.items(), unique=unique, session=self._session)
 
     def __ensure_indexes(self):
         if not object.__getattribute__(self, "_ensured_index"):
+            _disallow_transactions(self._session)
             self.__create_index(self._coll.files, _F_INDEX, False)
             self.__create_index(self._coll.chunks, _C_INDEX, True)
             object.__setattr__(self, "_ensured_index", True)
@@ -192,11 +223,12 @@ class GridIn(object):
     def abort(self):
         """Remove all chunks/files that may have been uploaded and close.
         """
-        self._coll.chunks.delete_many({"files_id": self._file['_id']})
-        self._coll.files.delete_one({"_id": self._file['_id']})
+        self._coll.chunks.delete_many(
+            {"files_id": self._file['_id']}, session=self._session)
+        self._coll.files.delete_one(
+            {"_id": self._file['_id']}, session=self._session)
         object.__setattr__(self, "_closed", True)
 
-
     @property
     def closed(self):
         """Is this file closed?
@@ -216,7 +248,7 @@ class GridIn(object):
                                     "Date that this file was uploaded.",
                                     closed_only=True)
     md5 = _grid_in_property("md5", "MD5 of the contents of this file "
-                            "(generated on the server).",
+                            "if an md5 sum was created.",
                             closed_only=True)
 
     def __getattr__(self, name):
@@ -241,10 +273,9 @@ class GridIn(object):
     def __flush_data(self, data):
         """Flush `data` to a chunk.
         """
-        # Ensure the index, even if there's nothing to write, so
-        # the filemd5 command always succeeds.
         self.__ensure_indexes()
-        self._file['md5'].update(data)
+        if 'md5' in self._file:
+            self._file['md5'].update(data)
 
         if not data:
             return
@@ -255,7 +286,7 @@ class GridIn(object):
                  "data": Binary(data)}
 
         try:
-            self._chunks.insert_one(chunk)
+            self._chunks.insert_one(chunk, session=self._session)
         except DuplicateKeyError:
             self._raise_file_exists(self._file['_id'])
         self._chunk_number += 1
@@ -274,11 +305,14 @@ class GridIn(object):
         try:
             self.__flush_buffer()
 
-            self._file['md5'] = self._file["md5"].hexdigest()
-            self._file["length"] = self._position
+            if "md5" in self._file:
+                self._file["md5"] = self._file["md5"].hexdigest()
+            # The GridFS spec says length SHOULD be an Int64.
+            self._file["length"] = Int64(self._position)
             self._file["uploadDate"] = datetime.datetime.utcnow()
 
-            return self._coll.files.insert_one(self._file)
+            return self._coll.files.insert_one(
+                self._file, session=self._session)
         except DuplicateKeyError:
             self._raise_file_exists(self._id)
 
@@ -296,6 +330,15 @@ class GridIn(object):
             self.__flush()
             object.__setattr__(self, "_closed", True)
 
+    def read(self, size=-1):
+        raise io.UnsupportedOperation('read')
+
+    def readable(self):
+        return False
+
+    def seekable(self):
+        return False
+
     def write(self, data):
         """Write data to the file. There is no return value.
 
@@ -363,6 +406,9 @@ class GridIn(object):
         for line in sequence:
             self.write(line)
 
+    def writeable(self):
+        return True
+
     def __enter__(self):
         """Support for the context manager protocol.
         """
@@ -382,7 +428,8 @@ class GridIn(object):
 class GridOut(object):
     """Class to read data out of GridFS.
     """
-    def __init__(self, root_collection, file_id=None, file_document=None):
+    def __init__(self, root_collection, file_id=None, file_document=None,
+                 session=None):
         """Read a file from GridFS
 
         Application developers should generally not need to
@@ -399,6 +446,17 @@ class GridOut(object):
           - `file_id` (optional): value of ``"_id"`` for the file to read
           - `file_document` (optional): file document from
             `root_collection.files`
+          - `session` (optional): a
+            :class:`~pymongo.client_session.ClientSession` to use for all
+            commands
+
+        .. versionchanged:: 3.8
+           For better performance and to better follow the GridFS spec,
+           :class:`GridOut` now uses a single cursor to read all the chunks in
+           the file.
+
+        .. versionchanged:: 3.6
+           Added ``session`` parameter.
 
         .. versionchanged:: 3.0
            Creating a GridOut does not immediately retrieve the file metadata
@@ -407,13 +465,18 @@ class GridOut(object):
         if not isinstance(root_collection, Collection):
             raise TypeError("root_collection must be an "
                             "instance of Collection")
+        _disallow_transactions(session)
+
+        root_collection = _clear_entity_type_registry(root_collection)
 
         self.__chunks = root_collection.chunks
         self.__files = root_collection.files
         self.__file_id = file_id
         self.__buffer = EMPTY
+        self.__chunk_iter = None
         self.__position = 0
         self._file = file_document
+        self._session = session
 
     _id = _grid_out_property("_id", "The ``'_id'`` value for this file.")
     filename = _grid_out_property("filename", "Name of this file.")
@@ -426,11 +489,13 @@ class GridOut(object):
     aliases = _grid_out_property("aliases", "List of aliases for this file.")
     metadata = _grid_out_property("metadata", "Metadata attached to this file.")
     md5 = _grid_out_property("md5", "MD5 of the contents of this file "
-                             "(generated on the server).")
+                             "if an md5 sum was created.")
 
     def _ensure_file(self):
         if not self._file:
-            self._file = self.__files.find_one({"_id": self.__file_id})
+            _disallow_transactions(self._session)
+            self._file = self.__files.find_one({"_id": self.__file_id},
+                                               session=self._session)
             if not self._file:
                 raise NoFile("no file in gridfs collection %r with _id %r" %
                              (self.__files, self.__file_id))
@@ -441,6 +506,9 @@ class GridOut(object):
             return self._file[name]
         raise AttributeError("GridOut object has no attribute '%s'" % name)
 
+    def readable(self):
+        return True
+
     def readchunk(self):
         """Reads a chunk at a time. If the current position is within a
         chunk the remainder of the chunk is returned.
@@ -453,11 +521,11 @@ class GridOut(object):
             chunk_data = self.__buffer
         elif self.__position < int(self.length):
             chunk_number = int((received + self.__position) / chunk_size)
-            chunk = self.__chunks.find_one({"files_id": self._id,
-                                            "n": chunk_number})
-            if not chunk:
-                raise CorruptGridFile("no chunk #%d" % chunk_number)
+            if self.__chunk_iter is None:
+                self.__chunk_iter = _GridOutChunkIterator(
+                    self, self.__chunks, self._session, chunk_number)
 
+            chunk = self.__chunk_iter.next()
             chunk_data = chunk["data"][self.__position % chunk_size:]
 
             if not chunk_data:
@@ -476,16 +544,21 @@ class GridOut(object):
 
         :Parameters:
           - `size` (optional): the number of bytes to read
+
+        .. versionchanged:: 3.8
+           This method now only checks for extra chunks after reading the
+           entire file. Previously, this method would check for extra chunks
+           on every call.
         """
         self._ensure_file()
 
-        if size == 0:
-            return EMPTY
-
         remainder = int(self.length) - self.__position
         if size < 0 or size > remainder:
             size = remainder
 
+        if size == 0:
+            return EMPTY
+
         received = 0
         data = StringIO()
         while received < size:
@@ -493,15 +566,12 @@ class GridOut(object):
             received += len(chunk_data)
             data.write(chunk_data)
 
-        # Detect extra chunks.
-        max_chunk_n = math.ceil(self.length / float(self.chunk_size))
-        chunk = self.__chunks.find_one({"files_id": self._id,
-                                        "n": {"$gte": max_chunk_n}})
-        # According to spec, ignore extra chunks if they are empty.
-        if chunk is not None and len(chunk['data']):
-            raise CorruptGridFile(
-                "Extra chunk found: expected %i chunks but found "
-                "chunk with n=%i" % (max_chunk_n, chunk['n']))
+        # Detect extra chunks after reading the entire file.
+        if size == remainder and self.__chunk_iter:
+            try:
+                self.__chunk_iter.next()
+            except StopIteration:
+                pass
 
         self.__position -= received - size
 
@@ -517,13 +587,13 @@ class GridOut(object):
         :Parameters:
          - `size` (optional): the maximum number of bytes to read
         """
-        if size == 0:
-            return b''
-
         remainder = int(self.length) - self.__position
         if size < 0 or size > remainder:
             size = remainder
 
+        if size == 0:
+            return EMPTY
+
         received = 0
         data = StringIO()
         while received < size:
@@ -574,8 +644,18 @@ class GridOut(object):
         if new_pos < 0:
             raise IOError(22, "Invalid value for `pos` - must be positive")
 
+        # Optimization, continue using the same buffer and chunk iterator.
+        if new_pos == self.__position:
+            return
+
         self.__position = new_pos
         self.__buffer = EMPTY
+        if self.__chunk_iter:
+            self.__chunk_iter.close()
+            self.__chunk_iter = None
+
+    def seekable(self):
+        return True
 
     def __iter__(self):
         """Return an iterator over all of this file's data.
@@ -584,12 +664,28 @@ class GridOut(object):
         :class:`str` (:class:`bytes` in python 3). This can be
         useful when serving files using a webserver that handles
         such an iterator efficiently.
+
+        .. note::
+           This is different from :py:class:`io.IOBase` which iterates over
+           *lines* in the file. Use :meth:`GridOut.readline` to read line by
+           line instead of chunk by chunk.
+
+        .. versionchanged:: 3.8
+           The iterator now raises :class:`CorruptGridFile` when encountering
+           any truncated, missing, or extra chunk in a file. The previous
+           behavior was to only raise :class:`CorruptGridFile` on a missing
+           chunk.
         """
-        return GridOutIterator(self, self.__chunks)
+        return GridOutIterator(self, self.__chunks, self._session)
 
     def close(self):
         """Make GridOut more generically file-like."""
-        pass
+        if self.__chunk_iter:
+            self.__chunk_iter.close()
+            self.__chunk_iter = None
+
+    def write(self, value):
+        raise io.UnsupportedOperation('write')
 
     def __enter__(self):
         """Makes it possible to use :class:`GridOut` files
@@ -601,28 +697,109 @@ class GridOut(object):
         """Makes it possible to use :class:`GridOut` files
         with the context manager protocol.
         """
+        self.close()
         return False
 
 
+class _GridOutChunkIterator(object):
+    """Iterates over a file's chunks using a single cursor.
+
+    Raises CorruptGridFile when encountering any truncated, missing, or extra
+    chunk in a file.
+    """
+    def __init__(self, grid_out, chunks, session, next_chunk):
+        self._id = grid_out._id
+        self._chunk_size = int(grid_out.chunk_size)
+        self._length = int(grid_out.length)
+        self._chunks = chunks
+        self._session = session
+        self._next_chunk = next_chunk
+        self._num_chunks = math.ceil(float(self._length) / self._chunk_size)
+        self._cursor = None
+
+    def expected_chunk_length(self, chunk_n):
+        if chunk_n < self._num_chunks - 1:
+            return self._chunk_size
+        return self._length - (self._chunk_size * (self._num_chunks - 1))
+
+    def __iter__(self):
+        return self
+
+    def _create_cursor(self):
+        filter = {"files_id": self._id}
+        if self._next_chunk > 0:
+            filter["n"] = {"$gte": self._next_chunk}
+        _disallow_transactions(self._session)
+        self._cursor = self._chunks.find(filter, sort=[("n", 1)],
+                                         session=self._session)
+
+    def _next_with_retry(self):
+        """Return the next chunk and retry once on CursorNotFound.
+
+        We retry on CursorNotFound to maintain backwards compatibility in
+        cases where two calls to read occur more than 10 minutes apart (the
+        server's default cursor timeout).
+        """
+        if self._cursor is None:
+            self._create_cursor()
+
+        try:
+            return self._cursor.next()
+        except CursorNotFound:
+            self._cursor.close()
+            self._create_cursor()
+            return self._cursor.next()
+
+    def next(self):
+        try:
+            chunk = self._next_with_retry()
+        except StopIteration:
+            if self._next_chunk >= self._num_chunks:
+                raise
+            raise CorruptGridFile("no chunk #%d" % self._next_chunk)
+
+        if chunk["n"] != self._next_chunk:
+            self.close()
+            raise CorruptGridFile(
+                "Missing chunk: expected chunk #%d but found "
+                "chunk with n=%d" % (self._next_chunk, chunk["n"]))
+
+        if chunk["n"] >= self._num_chunks:
+            # According to spec, ignore extra chunks if they are empty.
+            if len(chunk["data"]):
+                self.close()
+                raise CorruptGridFile(
+                    "Extra chunk found: expected %d chunks but found "
+                    "chunk with n=%d" % (self._num_chunks, chunk["n"]))
+
+        expected_length = self.expected_chunk_length(chunk["n"])
+        if len(chunk["data"]) != expected_length:
+            self.close()
+            raise CorruptGridFile(
+                "truncated chunk #%d: expected chunk length to be %d but "
+                "found chunk with length %d" % (
+                    chunk["n"], expected_length, len(chunk["data"])))
+
+        self._next_chunk += 1
+        return chunk
+
+    __next__ = next
+
+    def close(self):
+        if self._cursor:
+            self._cursor.close()
+            self._cursor = None
+
+
 class GridOutIterator(object):
-    def __init__(self, grid_out, chunks):
-        self.__id = grid_out._id
-        self.__chunks = chunks
-        self.__current_chunk = 0
-        self.__max_chunk = math.ceil(float(grid_out.length) /
-                                     grid_out.chunk_size)
+    def __init__(self, grid_out, chunks, session):
+        self.__chunk_iter = _GridOutChunkIterator(grid_out, chunks, session, 0)
 
     def __iter__(self):
         return self
 
     def next(self):
-        if self.__current_chunk >= self.__max_chunk:
-            raise StopIteration
-        chunk = self.__chunks.find_one({"files_id": self.__id,
-                                        "n": self.__current_chunk})
-        if not chunk:
-            raise CorruptGridFile("no chunk #%d" % self.__current_chunk)
-        self.__current_chunk += 1
+        chunk = self.__chunk_iter.next()
         return bytes(chunk["data"])
 
     __next__ = next
@@ -633,7 +810,8 @@ class GridOutCursor(Cursor):
     of an arbitrary query against the GridFS files collection.
     """
     def __init__(self, collection, filter=None, skip=0, limit=0,
-                 no_cursor_timeout=False, sort=None, batch_size=0):
+                 no_cursor_timeout=False, sort=None, batch_size=0,
+                 session=None):
         """Create a new cursor, similar to the normal
         :class:`~pymongo.cursor.Cursor`.
 
@@ -642,22 +820,27 @@ class GridOutCursor(Cursor):
 
         .. versionadded 2.7
 
-        .. mongodoc:: cursors
+        .. seealso:: The MongoDB documentation on `cursors <https://dochub.mongodb.org/core/cursors>`_.
         """
+        _disallow_transactions(session)
+        collection = _clear_entity_type_registry(collection)
+
         # Hold on to the base "fs" collection to create GridOut objects later.
         self.__root_collection = collection
 
         super(GridOutCursor, self).__init__(
             collection.files, filter, skip=skip, limit=limit,
             no_cursor_timeout=no_cursor_timeout, sort=sort,
-            batch_size=batch_size)
+            batch_size=batch_size, session=session)
 
     def next(self):
         """Get next GridOut object from cursor.
         """
+        _disallow_transactions(self.session)
         # Work around "super is not iterable" issue in Python 3.x
         next_file = super(GridOutCursor, self).next()
-        return GridOut(self.__root_collection, file_document=next_file)
+        return GridOut(self.__root_collection, file_document=next_file,
+                       session=self.session)
 
     __next__ = next
 
@@ -667,7 +850,7 @@ class GridOutCursor(Cursor):
     def remove_option(self, *args, **kwargs):
         raise NotImplementedError("Method does not exist for GridOutCursor")
 
-    def _clone_base(self):
+    def _clone_base(self, session):
         """Creates an empty GridOutCursor for information to be copied into.
         """
-        return GridOutCursor(self.__root_collection)
+        return GridOutCursor(self.__root_collection, session=session)
